import cv2
import numpy as np
from scipy import stats

from sdd_segmentation.sdd import sdd_threshold_selection
from bcanalyzer.image_processing.utils import invertByMask, getLargestContourThresh
from bcanalyzer.common.io import im_load


_img_cache = {
    'im_name': None,
    'raw_image_np': None
}


def mean_window(data, axis):
    res = np.sum(data, axis=axis)
    return res


def std_window(data, axis):
    res = np.std(data, axis=axis)
    return res

# Rolling 2D window for ND array


def roll(a,      # ND array
         b_shape,      # rolling 2D window array size
         dx=1,   # horizontal step, abscissa, number of columns
         dy=1):  # vertical step, ordinate, number of rows
    shape = a.shape[:-2] + \
        ((a.shape[-2] - b_shape[-2]) // dy + 1,) + \
        ((a.shape[-1] - b_shape[-1]) // dx + 1,) + \
        b_shape  # sausage-like shape with 2D cross-section
    strides = a.strides[:-2] + \
        (a.strides[-2] * dy,) + \
        (a.strides[-1] * dx,) + \
        a.strides[-2:]
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def sliding_window(data, win_shape, fcn, dx=1, dy=1):
    n = data.ndim  # number of dimensions
    # np.all over 2 dimensions of the rolling 2D window for 4D array
    result = fcn(roll(data, win_shape, dx, dy), axis=(n, n+1))
    return result


def tile_array(a, b0, b1):
    r, c = a.shape                                    # number of rows/columns
    rs, cs = a.strides                                # row/column strides
    # view a as larger 4D array
    x = np.lib.stride_tricks.as_strided(a, (r, b0, c, b1), (rs, 0, cs, 0))
    return x.reshape(r*b0, c*b1)                      # create new 2D array


def edge_density(img_np: np.array,
                 win_size: int,
                 win_step: int = 10,
                 canny_1: float = 41,
                 canny_2: float = 207) -> np.array:
    """Method is implemented algorithm for local edge density estimation, 
    proposed in "Sinitca, A. M., Kayumov, A. R., Zelenikhin, P. V., 
    Porfiriev, A. G., Kaplun, D. I., & Bogachev, M. I. (2023). 
    Segmentation of patchy areas in biomedical images based on local edge
    density estimation. Biomedical Signal Processing and Control, 79, 104189."

    https://www.sciencedirect.com/science/article/abs/pii/S1746809422006437

    Args:
        img_np (np.array): Gray scale image
        win_size (int): Size of averaging windows
        win_step (int, optional): Step for windows sliding. Defaults to 10.
        canny_1 (float, optional): 1st treshold for canny. Defaults to 41.
        canny_2 (float, optional): 2nd treshold for canny. Defaults to 207.

    Returns:
        np.array: Edge dencity map
    """
    dxy = win_step
    mid = cv2.Canny(img_np, canny_1, canny_2)

    result = sliding_window(mid, (win_size, win_size),
                            mean_window, dx=dxy, dy=dxy) // ((win_size*win_size))
    result = tile_array(result, dxy, dxy)
    h_pad = img_np.shape[0] - result.shape[0]
    w_pad = img_np.shape[1] - result.shape[1]
    result = np.pad(result, ((
        h_pad//2, h_pad//2+img_np.shape[0] % 2), (w_pad//2, w_pad//2+img_np.shape[1] % 2)), 'edge')
    return result


def process_image(img_path, params, largest_cnt_only=False, force_q_th=False):
    print(params)
    threshhold = params['thre']
    threshhold_abs = params['thre_abs']
    use_abs_threshhold = params['use_abs_threshhold']
    largest_cnt_only = params['is_single_object']
    do_bg_removing = params['do_bg_removing']
    do_otsu_thresholding = params['do_otsu_thresholding']

    win_size = int(params['win_size'])
    dxy = 10

    channel_r = params['channel_r']
    channel_g = params['channel_g']
    channel_b = params['channel_b']

    if _img_cache['im_name'] == img_path:
        img_np = _img_cache['raw_image_np'].copy()
    else:
        img_np = im_load(img_path)
        _img_cache['im_name'] = img_path
        _img_cache['raw_image_np'] = img_np.copy()

    if len(img_np.shape) == 3:
        img_np[:, :, 0] *= channel_b
        img_np[:, :, 1] *= channel_g
        img_np[:, :, 2] *= channel_r
        gray = cv2.cvtColor(img_np, cv2.COLOR_BGR2GRAY)
    elif len(img_np.shape) == 2:
        gray = img_np.copy()
        img_np = cv2.cvtColor(img_np, cv2.COLOR_GRAY2BGR)
    else:
        raise ValueError(
            f"Unsupported count of channels ({len(img_np.shape)})")

    if do_bg_removing:
        gray = gray.astype(float) / 255.0
        k_size = win_size*3
        bg = cv2.blur(gray, (k_size, k_size))
        gray = gray - bg

        gray -= gray.min()
        gray = (gray * 255).astype(np.uint8)

    mid = cv2.Canny(gray, 41, 207)

    result = sliding_window(mid, (win_size, win_size),
                            mean_window, dx=dxy, dy=dxy) // ((win_size*win_size))
    result = tile_array(result, dxy, dxy)
    h_pad = img_np.shape[0] - result.shape[0]
    w_pad = img_np.shape[1] - result.shape[1]
    result = np.pad(result, ((
        h_pad//2, h_pad//2+img_np.shape[0] % 2), (w_pad//2, w_pad//2+img_np.shape[1] % 2)), 'edge')

    th_perc = None
    if do_otsu_thresholding == "Otsu":
        th, img_thresh = cv2.threshold(
            np.uint8(result), 0, 255, cv2.THRESH_OTSU+cv2.THRESH_BINARY_INV)
        img_thresh = np.uint8(img_thresh)
    elif do_otsu_thresholding.startswith("SDD"):
        T = sdd_threshold_selection(result.astype(float), 15)
        if len(T) < 1:
            th = 0
        elif do_otsu_thresholding == "SDD lower":
            th = T[0]
        elif do_otsu_thresholding == "SDD median":
            th = T[len(T)//2]
        elif do_otsu_thresholding == "SDD upper":
            th = T[-1]

        img_thresh = np.where(result > th, 0, 1).astype(np.uint8)
    else:
        th = None
        if use_abs_threshhold and threshhold_abs is not None and not force_q_th:
            th = threshhold_abs
        else:
            th = np.percentile(result, threshhold)
            th_perc = threshhold
        img_thresh = np.where(result > th, 0, 1).astype(np.uint8)

    if th_perc is None:
        th_perc = stats.percentileofscore(result.flatten(), th, kind="weak")

    img_thresh = img_thresh[..., np.newaxis]
    #canny_thresh = cv2.bitwise_not(canny_thresh)
    if largest_cnt_only:
        img_thresh = getLargestContourThresh(img_thresh)

    preview = invertByMask(img_np, img_thresh)

    res_meta = {}
    res_meta["thre_abs"] = th
    res_meta["thre"] = th_perc
    print("res_meta", res_meta)
    return preview, img_thresh, res_meta




import sys
import os
import argparse
import ctypes
import logging
from bcanalyzer.common.const import APP_NAME
from bcanalyzer.gui.controllers.main_controller import MainController


def main():
    if os.name == 'nt':
        myappid = 'digiratory.bcanalyzer_app'  # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    """Application entry point"""
    parser = argparse.ArgumentParser(description="Microscopy image analyzer")

    try:
        MainController().run()
    except Exception as e:
        logging.exception(str(e))


if __name__ == "__main__":
    main()



class Singleton(type):
    """
    Singleton class.

    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Call function of Singleton.

        Parameters
        ----------
        args: Non-Keyword Arguments.
        kwargs: Keyword Arguments.

        Returns
        -------
        The instanced Singleton.

        """
        if cls not in cls._instances:
            cls._instances[cls] = super(
                Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bcanalyzer',
 'bcanalyzer.common',
 'bcanalyzer.gui.controllers',
 'bcanalyzer.gui.models',
 'bcanalyzer.gui.views',
 'bcanalyzer.gui.widgets',
 'bcanalyzer.image_processing']

package_data = \
{'': ['*'], 'bcanalyzer': ['gui/resources/*']}

install_requires = \
['PyQt5',
 'appdirs',
 'filetype',
 'numpy>=1.19.0,<2.0.0',
 'opencv-python-headless',
 'opencv-python>=4.5,<5.0',
 'pandas',
 'scipy',
 'sdd-segmentation']

entry_points = \
{'console_scripts': ['bcanalyzer = bcanalyzer.app:main']}

setup_kwargs = {
    'name': 'bio-cntrs-analyzer',
    'version': '0.1.11',
    'description': 'The semi-automatic segmentation and quantification of patchy areas in various biomedical images based on the assessment of their local edge densities.',
    'long_description': '# BCAnalyzer: Segmentation of patchy areas in biomedical images based on local edge density estimation\n\n## Installation\n\n### Download executable from release page\n \nWe provide a builded windows executable file for every release. You can find the latest whl or exe file on the [release page](https://gitlab.com/digiratory/biomedimaging/bcanalyzer/-/releases). Also, you can find all versions of windows execution files in the [archive](https://drive.digiratory.ru/d/s/nCQF5QKnlOoLyvoogzIMazY5YAOZDDmy/hY8ac89PJtnKtduC6w9CEkRtLVVRmZ4_-qbCgo_pvPwk).\n\n\n### Installation from binaries\n\n```bash\npip install -U bio-cntrs-analyzer\n```\n### Build exe from source\n\nFrom venv:\n\n```bash\npip install cx_Freeze\npython setup.py build\n```\n\n## User manual\n\nFor starting appliation evualate in command line (terminal) the next command:\n\n```bash\nbcanalyzer\n```\n\n![UserManualFigure](https://gitlab.com/digiratory/biomedimaging/bcanalyzer/-/raw/main/images/UserManualFigure.PNG) \n\nSoftware user interface outline: (A) list of images submitted for the analysis; (B) segmentation algorithm options, including (C) color channel import options; (D) selected image with on-the-fly visualization of the segmentation results; (E) segmentation algorithm controls for the online adjustment of its sensitivity and resolution; (F) file export menu.\n\nThe typical algorithm of user interaction with the software is as follows:\n\n* Selected images are imported by their drag-and-drop onto the program window. The image list appears in **A**.\n* Global algorithm options can be adjusted in **B** and color channels for the analysis selected in **C**.  \n* The image selected in the list **A** is displayed in the panel **D** with immediate visualization of the segmentation results (using default parameters during the first run).\n* Next the algorithm parameters (sensitivity and resolution) can be adjusted manually in **E**, given that the automated threshold option is disabled, although one can also apply the automated threshold selection for the first approach and then disable it in order to proceed with manual fine tuning. Segmentation results are visualized on-the-fly for direct user control. Of note, global options and color channel selection can be readjusted at this stage as well. Following necessary adjustments, the chosen algorithm parameters can be applied either to the entire imported image set, or solely to the currently analyzed image, with corresponding controls available in **E**.\n* Once the algorithm parameters are adjusted either for a single or for a few representative image(s), further processing and export can be performed as a fully automated procedure for the entire image set using file export options in **F**. Export options include visualizations of segmentation results, either as binary masks or as marked-up images similar to those appearing on the screen during the analysis, or both of them, as well as a *.csv table with summary statistics.\n\n# Dataset\n\nLink to downloading dataset:\n\nhttps://drive.digiratory.ru/d/s/mrbRyk4HyFbOIEANc6DhQGxhCFNgq3xI/Jqbcbsq_eE5Cf8-bnJaM3dXL6RI1v7d7-d74AOY4RLgk\n\nIf something goes wrong, please, write to amsinitca[at]etu.ru\n\n# Troubleshooting\n\n## qt.qpa.plugin: Could not load the Qt platform plugin "xcb" in "" even though it was found.\n\n### Problem\n\n```bash\nqt.qpa.plugin: Could not load the Qt platform plugin "xcb" in "" even though it was found.\nThis application failed to start because no Qt platform plugin could be initialized. Reinstalling the application may fix this problem.\nAvailable platform plugins are: eglfs, linuxfb, minimal, minimalegl, offscreen, vnc, wayland-egl, wayland, wayland-xcomposite-egl, wayland-xcomposite-glx, webgl, xcb.\nAborted (core dumped)\n```\n\n### Solution\n\n```bash\nsudo apt-get install libxcb-xinerama0\n```\n\n# Citation\n\nIf you find this project useful, please cite:\n\n```bib\n@article{SINITCA2023104189,\n    title = {Segmentation of patchy areas in biomedical images based on local edge density estimation},\n    journal = {Biomedical Signal Processing and Control},\n    volume = {79},\n    pages = {104189},\n    year = {2023},\n    issn = {1746-8094},\n    doi = {https://doi.org/10.1016/j.bspc.2022.104189},\n    url = {https://www.sciencedirect.com/science/article/pii/S1746809422006437},\n    author = {Aleksandr M. Sinitca and Airat R. Kayumov and Pavel V. Zelenikhin and Andrey G. Porfiriev and Dmitrii I. Kaplun and Mikhail I. Bogachev},\n}\n```\n\n',
    'author': 'Aleksandr Sinitca',
    'author_email': 'amsinitca@etu.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/digiratory/biomedimaging/bcanalyzer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
